import pycolmap

# def print_available_options(pipeline_options):
#     print("Available mapper_options for IncrementalPipelineOptions:")
#     for attr in dir(pipeline_options):
#         if not attr.startswith("__"):
#             print(f"  - {attr}")

def setup_incremental_pipeline_options():
    options = pycolmap.IncrementalPipelineOptions()
    
    # Set the global bundle adjustment tolerance, which is equivalent to the COLMAP ba_tolerance
    # Decreasing it speeds up bundle adjustment steps.
    options.ba_global_function_tolerance = 0.0000001  

    # Other bundle adjustment options
    options.ba_global_max_num_iterations = 100
    options.ba_global_max_refinements = 5
    options.ba_global_images_ratio = 1.1
    options.ba_global_points_ratio = 1.1
    options.ba_global_images_freq = 500
    options.ba_global_points_freq = 250000

    # Local bundle adjustment options
    options.ba_local_max_num_iterations = 25
    options.ba_local_function_tolerance = 0.0000001
    options.ba_local_max_refinements = 2
    options.ba_local_num_images = 6

    # Refinement options
    options.ba_refine_focal_length = True
    options.ba_refine_principal_point = False
    options.ba_refine_extra_params = True

    # Other relevant options
    options.min_num_matches = 15
    options.max_model_overlap = 20
    options.min_model_size = 10
    options.init_num_trials = 200
    options.extract_colors = True

    return options

def perform_bundle_adjustment(database_path, image_path, output_path):
    print("Starting bundle adjustment and mapping...")

    '''
    incremental_mapping() 
    The following argument types are supported:
        database_path: str, 
        image_path: str, 
        output_path: str, 
        options: pycolmap.IncrementalPipelineOptions = <pycolmap.IncrementalPipelineOptions object at 0x7f738ea48b30>, 
        input_path: str = '', 
        initial_image_pair_callback: Callable[[], None] = None, 
        next_image_callback: Callable[[], None] = None) -> Dict[int, pycolmap.Reconstruction]
    '''
    # Set up the pipeline options
    pipeline_options = setup_incremental_pipeline_options()
    
    # Print available pipeline_options for debugging
    # print_available_options(pipeline_options)

    # Run the incremental mapping
    try:
        reconstructions = pycolmap.incremental_mapping(
            database_path=database_path,
            image_path=image_path,
            output_path=output_path,
            options=pipeline_options
        )
        
        if reconstructions is None:
            print("Reconstruction failed.")
            return False
        
        return reconstructions
    except Exception as e:
        print(f"Bundle adjustment failed: {str(e)}")
        return False



#
# Copyright (C) 2023, Inria
# GRAPHDECO research group, https://team.inria.fr/graphdeco
# All rights reserved.
#
# This software is free for non-commercial, research and evaluation use
# under the terms of the LICENSE.md file.
#
# For inquiries contact  george.drettakis@inria.fr
#

import os
import torch
from random import randint
from .utils.loss_utils import l1_loss, ssim
from .gaussian_renderer import render, network_gui
from .scene import Scene, GaussianModel
from tqdm import tqdm

# Configure memory management
# torch.cuda.set_per_process_memory_fraction(0.9, 0)
# torch.cuda.set_max_split_size_mb(128)


def training(dataset, opt, pipe, saving_iterations, checkpoint_iterations, checkpoint, debug_from, ply_path):
    first_iter = 0
    dataset.model_path = ply_path
    dataset.source_path = os.path.join(ply_path, "data")
    gaussians = GaussianModel(dataset.sh_degree)
    scene = Scene(dataset, gaussians)
    gaussians.training_setup(opt)
    if checkpoint:
        (model_params, first_iter) = torch.load(checkpoint)
        gaussians.restore(model_params, opt)

    bg_color = [1, 1, 1] if dataset.white_background else [0, 0, 0]
    background = torch.tensor(bg_color, dtype=torch.float32, device="cuda")

    iter_start = torch.cuda.Event(enable_timing=True)
    iter_end = torch.cuda.Event(enable_timing=True)

    viewpoint_stack = None
    ema_loss_for_log = 0.0
    progress_bar = tqdm(range(first_iter, opt.iterations), desc="Training progress")
    first_iter += 1
    for iteration in range(first_iter, opt.iterations + 1):
        if network_gui.conn == None:
            network_gui.try_connect()
        while network_gui.conn != None:
            try:
                net_image_bytes = None
                custom_cam, do_training, pipe.convert_SHs_python, pipe.compute_cov3D_python, keep_alive, scaling_modifer = network_gui.receive()
                if custom_cam != None:
                    net_image = render(custom_cam, gaussians, pipe, background, scaling_modifer)["render"]
                    net_image_bytes = memoryview((torch.clamp(net_image, min=0, max=1.0) * 255).byte().permute(1, 2,
                                                                                                               0).contiguous().cpu().numpy())
                network_gui.send(net_image_bytes, dataset.source_path)
                if do_training and ((iteration < int(opt.iterations)) or not keep_alive):
                    break
            except Exception as e:
                network_gui.conn = None

        iter_start.record()

        gaussians.update_learning_rate(iteration)

        # Every 1000 its we increase the levels of SH up to a maximum degree
        if iteration % 1000 == 0:
            gaussians.oneupSHdegree()

        # Pick a random Camera
        if not viewpoint_stack:
            viewpoint_stack = scene.getTrainCameras().copy()
        viewpoint_cam = viewpoint_stack.pop(randint(0, len(viewpoint_stack) - 1))

        # Render
        if (iteration - 1) == debug_from:
            pipe.debug = True
        render_pkg = render(viewpoint_cam, gaussians, pipe, background)
        image, viewspace_point_tensor, visibility_filter, radii = render_pkg["render"], render_pkg["viewspace_points"], \
                                                                  render_pkg["visibility_filter"], render_pkg["radii"]

        # Loss
        gt_image = viewpoint_cam.original_image.cuda()
        Ll1 = l1_loss(image, gt_image)
        loss = (1.0 - opt.lambda_dssim) * Ll1 + opt.lambda_dssim * (1.0 - ssim(image, gt_image))
        loss.backward()

        iter_end.record()

        with torch.no_grad():
            # Progress bar
            ema_loss_for_log = 0.4 * loss.item() + 0.6 * ema_loss_for_log
            if iteration % 10 == 0:
                progress_bar.set_postfix({"Loss": f"{ema_loss_for_log:.{7}f}"})
                progress_bar.update(10)
            if iteration == opt.iterations:
                progress_bar.close()

            # Log and save
            if (iteration in saving_iterations) or (iteration == opt.iterations):
                print("\n[ITER {}] Saving Gaussians".format(iteration))
                scene.save(iteration)

            # Densification
            if iteration < opt.densify_until_iter:
                # Keep track of max radii in image-space for pruning
                gaussians.max_radii2D[visibility_filter] = torch.max(gaussians.max_radii2D[visibility_filter],
                                                                     radii[visibility_filter])
                gaussians.add_densification_stats(viewspace_point_tensor, visibility_filter)

                if iteration > opt.densify_from_iter and iteration % opt.densification_interval == 0:
                    size_threshold = 20 if iteration > opt.opacity_reset_interval else None
                    gaussians.densify_and_prune(opt.densify_grad_threshold, 0.005, scene.cameras_extent, size_threshold)

                if iteration % opt.opacity_reset_interval == 0 or (
                        dataset.white_background and iteration == opt.densify_from_iter):
                    gaussians.reset_opacity()

            # Optimizer step
            if iteration < opt.iterations:
                gaussians.optimizer.step()
                gaussians.optimizer.zero_grad(set_to_none=True)

            if (iteration in checkpoint_iterations):
                print("\n[ITER {}] Saving Checkpoint".format(iteration))
                torch.save((gaussians.capture(), iteration), scene.model_path + "/chkpnt" + str(iteration) + ".pth")





from setuptools import setup
from pathlib import Path


this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

packages = ['gsrecon']
requires = [
    "opencv-python",
    "pycolmap"
]

setup(name='gaussian-reconstruction',
	version='1.2.1',
	author='BerryChen',
    description="3D gaussian splatting recosntruction",
    packages=packages,
    long_description=long_description,
    long_description_content_type="text/markdown",
    package_dir={'requests': 'requests'},
    install_requires=requires,
    include_package_data=True,
)


# python converter.py -v C0010.MP4 -fps 1
import os
import argparse
import cv2
import math
import pycolmap
from tqdm import tqdm
#import distortion
from . import distortion
import time
import shutil
from pathlib import Path
from .train import training
from .arguments import ModelParams, ArgumentParser, OptimizationParams, PipelineParams


def measure_elapsed_time():
    start_time = time.time()

    def get_elapsed_time():
        end_time = time.time()
        elapsed_time = end_time - start_time
        hours = int(elapsed_time // 3600)
        minutes = int((elapsed_time % 3600) // 60)
        remaining_seconds = int(elapsed_time % 60)

        return f"Elapsed time: {hours} hrs {minutes} mins {remaining_seconds:.2f} secs"

    return get_elapsed_time


class VideoToPLYConverter:
    def __init__(self, method='pycolmap'):
        self.method = method

    # def convert_video_to_ply(self, video_path, fps, skip_ba):

    def convert_video_to_ply(self, video_path, fps):
        # Get the directory and filename without extension
        # directory = os.path.dirname(video_path)
        directory = os.getcwd()  # main dir
        filename = os.path.splitext(os.path.basename(video_path))[0]

        # Define the output file folder
        ply_path = Path(directory) / filename
        ply_path.mkdir(parents=True, exist_ok=True)  # filename

        if self.method == 'pycolmap':
            self._convert_using_pycolmap(video_path, ply_path, fps)
        # elif self.method ==
        else:
            raise ValueError("Invalid method.")

    def _extract_frames(self, image_dir, video_path, target_fps=None):
        # pre-define the maxmium frames extracted from the video
        max_frames = 500

        video = cv2.VideoCapture(video_path)

        if not video.isOpened():
            print("Error: Could not open video.")
            return

        original_fps = video.get(cv2.CAP_PROP_FPS)
        total_frames = int(video.get(cv2.CAP_PROP_FRAME_COUNT))
        video_duration = total_frames / original_fps
        # print(f"The video has an original fps of {original_fps} and total frames of {total_frames}")

        fps_to_use = target_fps if target_fps else round(max_frames / video_duration, 2)

        skip_frames = math.ceil(original_fps / fps_to_use)
        expected_frames = math.ceil(total_frames / skip_frames)

        # print(f"Extracting at a rate of {fps_to_use} and total frames of {expected_frames}")

        frame_count = 0
        saved_frame_idx = 0
        with tqdm(total=expected_frames, desc="Extracting frames", unit="frame") as pbar:
            while video.isOpened():
                ret, frame = video.read()
                if not ret:  # No more frames or error
                    break

                if frame_count % skip_frames == 0:
                    frame_filename = f"{image_dir}/{saved_frame_idx:04d}.jpg"
                    cv2.imwrite(frame_filename, frame)
                    saved_frame_idx += 1
                    pbar.update(1)

                frame_count += 1

        video.release()
        print(f"Extracted {saved_frame_idx} frames")

    def _convert_using_pycolmap(self, video_path, ply_path, fps):
        # Create necessary directories
        work_dir = os.path.join(ply_path, "data")  # filename/data
        if not os.path.exists(work_dir):
            os.mkdir(work_dir)

        image_dir = os.path.join(work_dir, "images")  # filename/data/images

        if os.path.exists(image_dir):
            shutil.rmtree(image_dir)
        os.makedirs(image_dir)

        # extract frames from input and save as images
        self._extract_frames(image_dir, video_path, fps)

        distorted_path = os.path.join(work_dir, "distorted")  # filename/data/distorted
        database_path = os.path.join(distorted_path, "database.db")  # filename/data/distorted/database.db
        sparse_path = os.path.join(distorted_path, "sparse")  # filename/data/distorted/sparse
        os.makedirs(distorted_path + "/sparse", exist_ok=True)

        if os.path.exists(database_path):
            os.remove(database_path)

        '''
        extract_features()
        The following argument types are supported:
            database_path: str, 
            image_path: str, 
            image_list: List[str] = [], 
            camera_mode: pycolmap.CameraMode = <CameraMode.AUTO: 0>, 
            camera_model: str = 'SIMPLE_RADIAL', 'PINHOLE'
            reader_options: pycolmap.ImageReaderOptions = <pycolmap.ImageReaderOptions object at 0x7f84459de730>, 
            sift_options: pycolmap.SiftExtractionOptions = <pycolmap.SiftExtractionOptions object at 0x7f84459de6b0>, 
            device: pycolmap.Device = <Device.auto: -1>) -> None
        '''
        pycolmap.extract_features(
            database_path,
            image_dir,
            camera_model='PINHOLE',
            sift_options={"max_num_features": 512, "gpu_index": '0'}
        )
        print("--------------------------- Extract features DONE ---------------------------")

        '''
        match_exhaustive()
        The following argument types are supported:
            database_path: str, 
            sift_options: pycolmap.SiftMatchingOptions = <pycolmap.SiftMatchingOptions object at 0x7ff0929d80f0>, 
            matching_options: pycolmap.ExhaustiveMatchingOptions = <pycolmap.ExhaustiveMatchingOptions 
        '''
        pycolmap.match_exhaustive(
            database_path,
            sift_options={"gpu_index": '0'}
        )
        print("--------------------------- Match features DONE ---------------------------")

        # incremental mapping
        maps = distortion.perform_bundle_adjustment(
            database_path,
            image_dir,
            distorted_path + '/sparse'
        )
        maps[0].write(work_dir)
        print("--------------------------- Incremental Mapping DONE ---------------------------")

        # Undistortion does not needed when camera model is SIMPLE_PINHOLE or PINHOLE
        # pycolmap.undistort_images(work_dir, sparse_path+'/0', image_dir)
        # print("--------------------------- Image Undistortion DONE ---------------------------")

        # If undistort_images() is called, manage the files into the right directory for training
        # self._manage_files(work_dir)
        # print("--------------------------- Manage files DONE ---------------------------")

        # resize images
        # for img_name in tqdm(os.listdir(image_dir), desc="Resizing images", unit="image"):
        #     img_path = os.path.join(image_dir, img_name)
        #     try:
        #         with Image.open(img_path) as img:
        #             img = img.resize((980, 545), Image.LANCZOS)
        #             img.save(img_path)  # Overwrite the larger image
        #     except Exception as e:
        #         print(f"Error processing {img_name}: {str(e)}")

        # print("--------------------------- Image resizing DONE ---------------------------")

        self._run_training(ply_path)
        print("--------------------------- Training process DONE ---------------------------")

    def _run_training(self, ply_path):
        print("--------------------------- Starting Training Process ---------------------------")

        # Define training parameters
        parser2 = ArgumentParser(description="Training script parameters")
        dataset = ModelParams(parser2)
        op = OptimizationParams(parser2)
        pp = PipelineParams(parser2)
        parser2.add_argument('--ip', type=str, default="127.0.0.1")
        parser2.add_argument('--port', type=int, default=6009)
        parser2.add_argument('--debug_from', type=int, default=-1)
        parser2.add_argument('--detect_anomaly', action='store_true', default=False)
        # parser2.add_argument("--testing_iterations", nargs="+", type=int, default=[30_000])
        # parser2.add_argument("--save_iterations", nargs="+", type=int, default=[1_000])
        parser2.add_argument("--save_iterations", nargs="+", type=int, default=[30_000, 70_000])
        parser2.add_argument("--quiet", action="store_true")
        parser2.add_argument("--checkpoint_iterations", nargs="+", type=int, default=[])
        parser2.add_argument("--start_checkpoint", type=str, default=None)
        parser2.add_argument("--saved_ply_path", type=str, default=ply_path)
        args2 = parser2.parse_args([])

        training(dataset.extract(args2),
                 op.extract(args2),
                 pp.extract(args2),
                 args2.save_iterations,
                 args2.checkpoint_iterations,
                 args2.start_checkpoint,
                 args2.debug_from,
                 args2.saved_ply_path)

        # # Assuming the training process generates a .ply file, return its path
        # result_path = os.path.join(dataset, "trained_model", "point_cloud.ply")

        # return result_path


# def main():
#     parser = argparse.ArgumentParser(description='Convert video to PLY file')
#     parser.add_argument('-v', '--video_path', type=str, help='Path to the input video file', required=True)
#     parser.add_argument('-fps', '--frame_rate', help="Extraction frame rate from the video", required=False, type=float)
#     # parser.add_argument('-ba', '--skip_ba', help="Skip incremental mapping and global bundle adjusement process",
#     #                     required=False, default=0, type=int)
#
#     args = parser.parse_args()
#
#     timer = measure_elapsed_time()
#
#     converter = VideoToPLYConverter(method='pycolmap')
#     try:
#         # output_ply = converter.convert_video_to_ply(args.video_path, args.frame_rate, args.skip_ba)
#         output_ply = converter.convert_video_to_ply(args.video_path, args.frame_rate)
#         # print(f"Conversion successful. PLY file created at: {output_ply}")
#     except Exception as e:
#         print(f"Error: {str(e)}")
#
#     print(timer())
#
#
# if __name__ == "__main__":
#     main()

